/* 
 * disk IO performance measurement 
 *
 * Written by: Kevin Hilman
 *
 * NOTES: 
 *  - assmes /dev/hda1 is minix filesystem
 *  - assumes /tmp_mnt directory exists on disk
 *  - bytes/second calculations assume 256MHz core speed
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#ifdef _ETI_
#include <eti/mm.h>
#define USE_FCNT
#endif

#ifndef USE_FCNT
#include <sys/times.h>
#endif

#include <sys/mount.h>

#define BUF_SIZE (512<<10)
#define NUM_LOOPS 16

int file_read_test(char *filename) {
  int fd, result;
  unsigned char *buf;
  unsigned long i, j;
#ifdef USE_FCNT
  unsigned long fcnt, diff;
#else
  struct tms t0, t1;
  unsigned long ticks;
#endif

  buf = (unsigned char *)malloc(BUF_SIZE);
  if (buf == NULL) {
    perror("read_test: malloc() failed");
    return -1;
  }

  fd = open(filename, O_RDONLY);
  if (fd < 0) {
    printf("read_test: open() failed: %d", errno);
    return -1;
  }

  for(j=0; j<NUM_LOOPS; j++) {
#ifdef USE_FCNT
    fcnt = hmpv_read_fcnt_cl0();
#else
    times(&t0);
#endif
    /* Read */
    result = read(fd, buf, BUF_SIZE);
#ifdef USE_FCNT
    diff = hmpv_read_fcnt_cl0() - fcnt;
#else
    times(&t1);
    ticks = t1.tms_stime - t0.tms_stime;
#endif

    if (result < 0) {
      printf("read_test: read() failed: %d", result);
      return -1;
    }
    else {
#ifdef USE_FCNT
      printf("read_test : %d Kbytes in %d Kcycles (%.2e bytes/sec).\n", 
	     BUF_SIZE>>10, diff>>10, BUF_SIZE/(double)diff * 256e6);
#else
      printf("read_test: read %d bytes in %d ticks.\n",
	     BUF_SIZE, ticks);
#endif
    }

    /* Compare */
    for(i=0; i<BUF_SIZE; i++) {
      if (buf[i] != (unsigned char)i) {
	printf("read_test: ERROR: wrote 0x%x, read 0x%x\n", i, buf[i]);
      }
    }
  }

  result = close(fd);
  if (result < 0) {
    printf("read_test: close() failed: %d", result);
  }

  free(buf);
  return result;
}

int file_write_test(char *filename) {
  int fd, result;
  unsigned char *buf;
  struct stat st;
  unsigned long i, j, open_flags;
#ifdef USE_FCNT
  unsigned long fcnt, diff;
#else
  struct tms t0, t1;
  unsigned long ticks;
#endif

  buf = (unsigned char *)malloc(BUF_SIZE);
  if (buf == NULL) {
    perror("malloc() failed");
    return -1;
  }
  for(i=0; i<BUF_SIZE; i++)
    buf[i] = (unsigned char)i;

  result = stat(filename, &st);
  if (result < 0) {
    printf("write_test: stat() failed: %d\n", result);
  }
  else {
/*      printf("write_test: file '%s' exists: size=%d bytes.  Deleting.\n",  */
/*  	   filename, st.st_size); */

    result = unlink(filename);
    if (result < 0) {
      printf("write_test: unlink() failed: %d\n", result);
    }
  }

  open_flags = O_WRONLY | O_CREAT | O_TRUNC;
/*  open_flags |= O_SYNC; */
  fd = open(filename, open_flags, 00644);
  if (fd < 0) {
    printf("write_test: open() failed: %d\n", errno);
    return -1;
  }

  for(j=0; j<NUM_LOOPS; j++) {
#ifdef USE_FCNT
    fcnt = hmpv_read_fcnt_cl0();
#else
    times(&t0);
#endif
    result = write(fd, buf, BUF_SIZE);
#ifdef USE_FCNT
    diff = hmpv_read_fcnt_cl0() - fcnt;
#else
    times(&t1);
    ticks = t1.tms_stime - t0.tms_stime;
#endif
    if (result < 0) {
      printf("write_test: write() failed: %d\n", result);
      return -1;
    }
    else {
#ifdef USE_FCNT
      printf("write_test: %d Kbytes in %d Kcycles (%.2e bytes/sec).\n", 
	     BUF_SIZE>>10, diff>>10, BUF_SIZE/(double)diff * 256e6);
#else
      printf("write_test: wrote %d bytes in %d ticks.\n",
	     BUF_SIZE, ticks);
#endif
    }
  }

  result = close(fd);
  if (result < 0) {
    perror("close() failed");
  }

  free(buf);
  return result;
}

int main(int argc, char *argv[], char *envp[]) {
  int result;
  char *filename = "/tmp_mnt/testfile.bin";
  unsigned long rwflag = (0xc0ed << 16) | 0x0;
  char *options = "debug,check";
  clock_t start, end;

  printf("fileIO test\n");
  start = clock();
  result = mount("/dev/hda1", "/tmp_mnt", "minix", rwflag, options);
  if (result < 0) {
    printf("mount() returned %d\n", result);
    return -1;
  }

  result = file_write_test(filename);
  if (result == 0)
    file_read_test(filename);

  result = umount("/tmp_mnt");
  if (result < 0) {
    printf("umount() returned %d\n", result);
    return -1;
  }

  end = clock();
  printf("fileIO test done: used %d ticks\n", end-start);
  return 0;
}
