/*
 * In RAID Engine, reserved sectors are used for preserving the information
 * (or structures) of the disk array and disk itself, which is located at the
 * first six sectors in the last cylinder of each disk.  Normally the first
 * four sectors will be accessed during initializing RAID or other read/write
 * reserved sector operation; the fifth sector is just used to save array
 * boot information used by the BIOS.
 */

#define SIGNATURE_SIZE	24
#define SERIAL_SIZE	8
#define bit(x)		(1 << x)

/* id  Structure
 *	Note:
 * 1.	WFormat:  here is 0x00020000;
 * 2.	SIGNATURE_SIZE=24;
 * 3.	SERIAL_SIZE =8;
 *	bSerial[0]-bSerial[5]:	put random digits got from timestamp.
 *	bSerial[6]-bSerial[7]:	disk Channel ID and Master/slave ID.
 * 4.	bDate [6]:		put random digits got from timestamp.
 */

typedef struct id {
	__u8 bSignature[SIGNATURE_SIZE];	/* Promise Technology, Inc. */
	__u32 wFormat;
	__u8 bSerial[SERIAL_SIZE];
	__u8 bDate[6];				/* 2 U8 days + 4 U8 ticks */
	__u8 bReserved[6];
} ID, buf_space *bpID;

/* Logical disk structure */
typedef struct stream {
	BOOL blFlag;			/* see  BLOCK blFlag */
	__u8 bIndex;			/* physical disk ID */
	__u8 bCh;			/* channel ID for logical disk */
	__u8 bID;			/* bID: 0 master; 1 slave  */
	__u8 bSerial[SERIAL_SIZE];	/* serial number of logical disk */
} STREAM, buf_space *bpSTREAM;

/* The part of array structure */
 * Note:
 *  1. hGeneration: increase 1 every time when reserved sector updating;
 *  2. blFlag:
 *	blVALID_LUN		bit(0) valid array
 *	blWORKING_LUN		bit(1) working array
 *	blINITIALIZED_LUN	bit(2) initialized array
 *	blARY_READY		bit(3) array status is ready
 *	blRECOVERING_LUN	bit(4) array is in degraded mode
 *	blMARKED_LUN		bit(5) need update array data to reserved sector
 *	bReserved		bit(6)
 *	blBOUNDARY		bit(7) array size should align to 1M boundary
 *
 *  3. bLogicalWith:
 *		for striping	bLogicalWidth = bWidth
 *		for mirroring	bLogicalWidth = 1 (two disks)
 *				bLogicalWidth = 2(Four disks)
 *  4. bStripeBlockPower:
 *		power of 2 for stripe block size, the stripe block
 *		size can be  calculated from it.
 */

typedef struct lun {
	__u16 hGeneration;
	BOOL blFlag;		/* see flag definition */
	BOOL blRAID;		/* Raid mode */
#define blMIRROR	bit(0)	/* blRAID */
#define blPARITY	bit(1)
#define blDISTRIBUTED	bit(2)
#define blSPAN		bit(3)
#define blRAID0		0
#define blRAID1		blMIRROR
#define blRAID4		blPARITY
#define blRAID5		blDISTRIBUTED
	__u8 bWidth;		/* the number of all disks in the array */
	__u8 bStripeBlockPower;	/* power of 2 */
	__u8 bLogicalWidth;
	__u8 bID;		/* array ID */
	__u32 wSize;		/* total count of sector */
	__u16 hCyl;		/* # of cyl */
	__u8 bHd;		/* # of hd per cyl */
	__u8 bSctr;		/* # of sctr per hd */
	__u8 bSerial [SERIAL_SIZE];
#define MAX_STREAM_PER_ARRAY		0x08
	STREAM strm [MAX_STREAM_PER_ARRAY];	/* logical disk structure */
} LUN, buf_space *bpLUN;

/* physical block (disk)  structure */
 * Note:
 * 1.BLOCK  blFlag :
 * blVALID_BLK          bit(0)  valid disk
 * blWORKING_BLK        bit(1)  indicate this disk is a working disk
 * blASSIGNED_BLK       bit(2)  assigned to an array
 * blSPARE_BLK          bit(3)  not used anymore
 * blDUPLICATE_BLK      bit(4)  not used anymore
 * blREDIRECT_BLK       bit(5)  redirect to rebuild procedure
 * blMARK_FOR_DOWN      bit(6)  marked as down (should be set down)
 * blNEED_HOT_SPARE     bit(6)  should be put to a array
 * blMARK_FOR_UP        bit(7)  marked as up (should be put to a array)
 *
 * Normally, blFlag should be 0x7:  blVALID_BLK+blWORKING_BLK+blASSINGED_BLK
 * for free disk blFlag should be 0x3: blVALID_BLK+blWORKING_BLK
 */
typedef struct block {
	BOOL blFlag;		/* see block flag definition */
	__u8 bStream;		/* logical disk ID */
	__u8 bCh;		/* channel ID of the physical disk */
	__u8 bID;		/* 0: master; 1: slave  */
	__u8 bSerial[SERIAL_SIZE];	/*  SERIAL_SIZE = 8 */
	__u32 wStartLBA;	/* disk watermark for Synchronization  */
	__u32 wSize;		/* total count of sectors */
	__u32 wWatermark;	/* disk watermark for Rebuilding */
} BLOCK, buf_space *bpBLOCK, dat_space *dpBLOCK;

/* assoc  structure */
typedef struct assoc {
	__u8 blFlag;
	__u8 bReserved[3];
	BLOCK blk;		/* physical disk structure */
	LUN lun;		/* the part of array structure */
} ASSOC, buf_space *bpASSOC;

typedef struct array {
	__u16 hGeneration;
	BOOL blFlag;
	BOOL blRAID;			/* RAID mode  */
	__u8 bWidth;			/* the number of disks in the array */
	__u8 bStripeBlockPower;		/* in power of 2 sctrs */
	__u8 bLogicalWidth;
	__u8 bID;			/* array ID  */
	__u32 wSize;			/* # of sectors */
	__u16 hCyl;			/* # of cyl */
	__u8 bHd;			/* # of hd per cyl */
	__u8 bSctr;			/* # of sctr per hd */
	__u8 bSerial[SERIAL_SIZE];
	__u32 wWatermark;
	__u16 hStripeBlockSize;		/* Stripe Block Size in sector number */
	__u16 hStripeSize;		/* Stripe Size in sector number */
	STRM strm[MAX_STREAM_PER_ARRAY];	/* index into block database */
	__u8 bInit;	/* Initializing Flag when creating a new array */
	__u8 bReserved;
	__u8 Name[64];	/* array name, just used for SuperTrak  */
	__u32 ArrayTid;	/* just used for SuperTrak/UltraTrak  */
} ARY, dat_space *dpARY;

/* Reserved Sector Structure */
typedef struct disk {
	ID    id;				/* sector 0 */
	__u8  bRestOfSector0[512-sizeof(ID)];
	ASSOC links[MAX_BLOCK_PER_DISK];	/* sector 1 */
	__u8  bRestOfSector1[512-sizeof(ASSOC)];
	__u8  bRestOfSector2[512];		/* sector 2 */
	__u8  bRestOfSector3[512-4];		/* sector 3 */
	__u32 wChecksum;
	__u8  bRestOfSector4[512];		/* sector 4 */
	__u8  bRestOfSector5[512];		/* sector 5 */
} DISK, buf_space *bpDISK;

